EXPORTS

g_mime_application_pkcs7_mime_decrypt
g_mime_application_pkcs7_mime_encrypt
g_mime_application_pkcs7_mime_get_smime_type
g_mime_application_pkcs7_mime_get_type
g_mime_application_pkcs7_mime_new
g_mime_application_pkcs7_mime_sign
g_mime_application_pkcs7_mime_verify
g_mime_certificate_get_created
g_mime_certificate_get_digest_algo
g_mime_certificate_get_email
g_mime_certificate_get_expires
g_mime_certificate_get_fingerprint
g_mime_certificate_get_id_validity
g_mime_certificate_get_issuer_name
g_mime_certificate_get_issuer_serial
g_mime_certificate_get_key_id
g_mime_certificate_get_name
g_mime_certificate_get_pubkey_algo
g_mime_certificate_get_trust
g_mime_certificate_get_type
g_mime_certificate_get_user_id
g_mime_certificate_list_add
g_mime_certificate_list_clear
g_mime_certificate_list_contains
g_mime_certificate_list_get_certificate
g_mime_certificate_list_get_type
g_mime_certificate_list_index_of
g_mime_certificate_list_insert
g_mime_certificate_list_length
g_mime_certificate_list_new
g_mime_certificate_list_remove
g_mime_certificate_list_remove_at
g_mime_certificate_list_set_certificate
g_mime_certificate_new
g_mime_certificate_set_created
g_mime_certificate_set_digest_algo
g_mime_certificate_set_email
g_mime_certificate_set_expires
g_mime_certificate_set_fingerprint
g_mime_certificate_set_id_validity
g_mime_certificate_set_issuer_name
g_mime_certificate_set_issuer_serial
g_mime_certificate_set_key_id
g_mime_certificate_set_name
g_mime_certificate_set_pubkey_algo
g_mime_certificate_set_trust
g_mime_certificate_set_user_id
g_mime_charset_best
g_mime_charset_best_name
g_mime_charset_can_encode
g_mime_charset_canon_name
g_mime_charset_iconv_name
g_mime_charset_init
g_mime_charset_iso_to_windows
g_mime_charset_language
g_mime_charset_locale_name
g_mime_charset_map_init
g_mime_charset_map_shutdown
g_mime_charset_name
g_mime_charset_step
g_mime_check_version
g_mime_content_disposition_encode
g_mime_content_disposition_get_disposition
g_mime_content_disposition_get_parameter
g_mime_content_disposition_get_parameters
g_mime_content_disposition_get_type
g_mime_content_disposition_is_attachment
g_mime_content_disposition_new
g_mime_content_disposition_parse
g_mime_content_disposition_set_disposition
g_mime_content_disposition_set_parameter
g_mime_content_encoding_from_string
g_mime_content_encoding_to_string
g_mime_content_type_encode
g_mime_content_type_get_media_subtype
g_mime_content_type_get_media_type
g_mime_content_type_get_mime_type
g_mime_content_type_get_parameter
g_mime_content_type_get_parameters
g_mime_content_type_get_type
g_mime_content_type_is_type
g_mime_content_type_new
g_mime_content_type_parse
g_mime_content_type_set_media_subtype
g_mime_content_type_set_media_type
g_mime_content_type_set_parameter
g_mime_crypto_context_decrypt
g_mime_crypto_context_digest_id
g_mime_crypto_context_digest_name
g_mime_crypto_context_encrypt
g_mime_crypto_context_export_keys
g_mime_crypto_context_get_encryption_protocol
g_mime_crypto_context_get_key_exchange_protocol
g_mime_crypto_context_get_request_password
g_mime_crypto_context_get_signature_protocol
g_mime_crypto_context_get_type
g_mime_crypto_context_import_keys
g_mime_crypto_context_new
g_mime_crypto_context_register
g_mime_crypto_context_set_request_password
g_mime_crypto_context_shutdown
g_mime_crypto_context_sign
g_mime_crypto_context_verify
g_mime_data_wrapper_get_encoding
g_mime_data_wrapper_get_stream
g_mime_data_wrapper_get_type
g_mime_data_wrapper_new
g_mime_data_wrapper_new_with_stream
g_mime_data_wrapper_set_encoding
g_mime_data_wrapper_set_stream
g_mime_data_wrapper_write_to_stream
g_mime_decrypt_result_get_cipher
g_mime_decrypt_result_get_mdc
g_mime_decrypt_result_get_recipients
g_mime_decrypt_result_get_session_key
g_mime_decrypt_result_get_signatures
g_mime_decrypt_result_get_type
g_mime_decrypt_result_new
g_mime_decrypt_result_set_cipher
g_mime_decrypt_result_set_mdc
g_mime_decrypt_result_set_recipients
g_mime_decrypt_result_set_session_key
g_mime_decrypt_result_set_signatures
g_mime_encoding_base64_decode_step
g_mime_encoding_base64_encode_close
g_mime_encoding_base64_encode_step
g_mime_encoding_flush
g_mime_encoding_init_decode
g_mime_encoding_init_encode
g_mime_encoding_outlen
g_mime_encoding_quoted_decode_step
g_mime_encoding_quoted_encode_close
g_mime_encoding_quoted_encode_step
g_mime_encoding_reset
g_mime_encoding_step
g_mime_encoding_uudecode_step
g_mime_encoding_uuencode_close
g_mime_encoding_uuencode_step
g_mime_filter_backup
g_mime_filter_basic_get_type
g_mime_filter_basic_new
g_mime_filter_best_charset
g_mime_filter_best_encoding
g_mime_filter_best_get_type
g_mime_filter_best_new
g_mime_filter_charset_get_type
g_mime_filter_charset_new
g_mime_filter_checksum_get_digest
g_mime_filter_checksum_get_string
g_mime_filter_checksum_get_type
g_mime_filter_checksum_new
g_mime_filter_complete
g_mime_filter_copy
g_mime_filter_dos2unix_get_type
g_mime_filter_dos2unix_new
g_mime_filter_enriched_get_type
g_mime_filter_enriched_new
g_mime_filter_filter
g_mime_filter_from_get_type
g_mime_filter_from_new
g_mime_filter_get_type
g_mime_filter_gzip_get_comment
g_mime_filter_gzip_get_filename
g_mime_filter_gzip_get_type
g_mime_filter_gzip_new
g_mime_filter_gzip_set_comment
g_mime_filter_gzip_set_filename
g_mime_filter_html_get_type
g_mime_filter_html_new
g_mime_filter_openpgp_new
g_mime_filter_openpgp_get_data_type
g_mime_filter_openpgp_get_begin_offset
g_mime_filter_openpgp_get_end_offset
g_mime_filter_openpgp_get_type
g_mime_filter_reset
g_mime_filter_set_size
g_mime_filter_smtp_data_get_type
g_mime_filter_smtp_data_new
g_mime_filter_strip_get_type
g_mime_filter_strip_new
g_mime_filter_unix2dos_get_type
g_mime_filter_unix2dos_new
g_mime_filter_windows_get_type
g_mime_filter_windows_is_windows_charset
g_mime_filter_windows_new
g_mime_filter_windows_real_charset
g_mime_filter_yenc_get_crc
g_mime_filter_yenc_get_pcrc
g_mime_filter_yenc_get_type
g_mime_filter_yenc_new
g_mime_filter_yenc_set_crc
g_mime_filter_yenc_set_state
g_mime_format_options_add_hidden_header
g_mime_format_options_clear_hidden_headers
g_mime_format_options_clone
g_mime_format_options_create_newline_filter
g_mime_format_options_free
g_mime_format_options_get_default
g_mime_format_options_get_newline
g_mime_format_options_get_newline_format
g_mime_format_options_get_param_encoding_method
g_mime_format_options_get_type
g_mime_format_options_is_hidden_header
g_mime_format_options_new
g_mime_format_options_remove_hidden_header
g_mime_format_options_set_newline_format
g_mime_format_options_set_param_encoding_method
g_mime_gpg_context_get_type
g_mime_gpg_context_new
g_mime_header_format_addrlist
g_mime_header_format_content_disposition
g_mime_header_format_content_type
g_mime_header_format_default
g_mime_header_format_message_id
g_mime_header_format_received
g_mime_header_format_references
g_mime_header_get_name
g_mime_header_get_offset
g_mime_header_get_raw_name
g_mime_header_get_raw_value
g_mime_header_get_type
g_mime_header_get_value
g_mime_header_list_append
g_mime_header_list_clear
g_mime_header_list_contains
g_mime_header_list_get_count
g_mime_header_list_get_header
g_mime_header_list_get_header_at
g_mime_header_list_get_type
g_mime_header_list_new
g_mime_header_list_prepend
g_mime_header_list_remove
g_mime_header_list_remove_at
g_mime_header_list_set
g_mime_header_list_to_string
g_mime_header_list_write_to_stream
g_mime_header_set_raw_value
g_mime_header_set_value
g_mime_header_write_to_stream
g_mime_iconv_close
g_mime_iconv_locale_to_utf8
g_mime_iconv_locale_to_utf8_length
g_mime_iconv_open
g_mime_iconv_strdup
g_mime_iconv_strndup
g_mime_iconv_utf8_to_locale
g_mime_iconv_utf8_to_locale_length
g_mime_init
g_mime_locale_charset
g_mime_locale_language
g_mime_message_add_mailbox
g_mime_message_foreach
g_mime_message_get_addresses
g_mime_message_get_all_recipients
g_mime_message_get_autocrypt_gossip_headers
g_mime_message_get_autocrypt_gossip_headers_from_inner_part
g_mime_message_get_autocrypt_header
g_mime_message_get_bcc
g_mime_message_get_body
g_mime_message_get_cc
g_mime_message_get_date
g_mime_message_get_from
g_mime_message_get_message_id
g_mime_message_get_mime_part
g_mime_message_get_reply_to
g_mime_message_get_sender
g_mime_message_get_subject
g_mime_message_get_to
g_mime_message_get_type
g_mime_message_new
g_mime_message_part_get_message
g_mime_message_part_get_type
g_mime_message_part_new
g_mime_message_part_new_with_message
g_mime_message_part_set_message
g_mime_message_partial_get_id
g_mime_message_partial_get_number
g_mime_message_partial_get_total
g_mime_message_partial_get_type
g_mime_message_partial_new
g_mime_message_partial_reconstruct_message
g_mime_message_partial_split_message
g_mime_message_set_date
g_mime_message_set_message_id
g_mime_message_set_mime_part
g_mime_message_set_subject
g_mime_multipart_add
g_mime_multipart_clear
g_mime_multipart_contains
g_mime_multipart_encrypted_decrypt
g_mime_multipart_encrypted_encrypt
g_mime_multipart_encrypted_get_type
g_mime_multipart_encrypted_new
g_mime_multipart_foreach
g_mime_multipart_get_boundary
g_mime_multipart_get_count
g_mime_multipart_get_epilogue
g_mime_multipart_get_part
g_mime_multipart_get_prologue
g_mime_multipart_get_subpart_from_content_id
g_mime_multipart_get_type
g_mime_multipart_index_of
g_mime_multipart_insert
g_mime_multipart_new
g_mime_multipart_new_with_subtype
g_mime_multipart_remove
g_mime_multipart_remove_at
g_mime_multipart_replace
g_mime_multipart_set_boundary
g_mime_multipart_set_epilogue
g_mime_multipart_set_prologue
g_mime_multipart_signed_get_type
g_mime_multipart_signed_new
g_mime_multipart_signed_sign
g_mime_multipart_signed_verify
g_mime_object_append_header
g_mime_object_encode
g_mime_object_get_content_disposition
g_mime_object_get_content_disposition_parameter
g_mime_object_get_content_id
g_mime_object_get_content_type
g_mime_object_get_content_type_parameter
g_mime_object_get_disposition
g_mime_object_get_header
g_mime_object_get_header_list
g_mime_object_get_headers
g_mime_object_get_type
g_mime_object_new
g_mime_object_new_type
g_mime_object_prepend_header
g_mime_object_register_type
g_mime_object_remove_header
g_mime_object_set_content_disposition
g_mime_object_set_content_disposition_parameter
g_mime_object_set_content_id
g_mime_object_set_content_type
g_mime_object_set_content_type_parameter
g_mime_object_set_disposition
g_mime_object_set_header
g_mime_object_to_string
g_mime_object_write_to_stream
g_mime_param_get_charset
g_mime_param_get_encoding_method
g_mime_param_get_lang
g_mime_param_get_name
g_mime_param_get_type
g_mime_param_get_value
g_mime_param_list_clear
g_mime_param_list_encode
g_mime_param_list_get_parameter
g_mime_param_list_get_parameter_at
g_mime_param_list_get_type
g_mime_param_list_length
g_mime_param_list_new
g_mime_param_list_parse
g_mime_param_list_remove
g_mime_param_list_remove_at
g_mime_param_list_set_parameter
g_mime_param_set_charset
g_mime_param_set_encoding_method
g_mime_param_set_lang
g_mime_param_set_value
g_mime_parser_construct_message
g_mime_parser_construct_part
g_mime_parser_eos
g_mime_parser_get_format
g_mime_parser_get_headers_begin
g_mime_parser_get_headers_end
g_mime_parser_get_mbox_marker
g_mime_parser_get_mbox_marker_offset
g_mime_parser_get_persist_stream
g_mime_parser_get_respect_content_length
g_mime_parser_get_type
g_mime_parser_init_with_stream
g_mime_parser_new
g_mime_parser_new_with_stream
g_mime_parser_options_clone
g_mime_parser_options_free
g_mime_parser_options_get_address_compliance_mode
g_mime_parser_options_get_allow_addresses_without_domain
g_mime_parser_options_get_default
g_mime_parser_options_get_fallback_charsets
g_mime_parser_options_get_parameter_compliance_mode
g_mime_parser_options_get_rfc2047_compliance_mode
g_mime_parser_options_get_type
g_mime_parser_options_get_warning_callback
g_mime_parser_options_new
g_mime_parser_options_set_address_compliance_mode
g_mime_parser_options_set_allow_addresses_without_domain
g_mime_parser_options_set_fallback_charsets
g_mime_parser_options_set_parameter_compliance_mode
g_mime_parser_options_set_rfc2047_compliance_mode
g_mime_parser_options_set_warning_callback
g_mime_parser_set_format
g_mime_parser_set_header_regex
g_mime_parser_set_persist_stream
g_mime_parser_set_respect_content_length
g_mime_parser_tell
g_mime_part_get_best_content_encoding
g_mime_part_get_content
g_mime_part_get_content_description
g_mime_part_get_content_encoding
g_mime_part_get_content_id
g_mime_part_get_content_location
g_mime_part_get_content_md5
g_mime_part_get_filename
g_mime_part_get_openpgp_data
g_mime_part_get_type
g_mime_part_is_attachment
g_mime_part_iter_clone
g_mime_part_iter_free
g_mime_part_iter_get_current
g_mime_part_iter_get_parent
g_mime_part_iter_get_path
g_mime_part_iter_get_toplevel
g_mime_part_iter_get_type
g_mime_part_iter_is_valid
g_mime_part_iter_jump_to
g_mime_part_iter_new
g_mime_part_iter_next
g_mime_part_iter_prev
g_mime_part_iter_remove
g_mime_part_iter_replace
g_mime_part_iter_reset
g_mime_part_new
g_mime_part_new_with_type
g_mime_part_openpgp_decrypt
g_mime_part_openpgp_encrypt
g_mime_part_openpgp_sign
g_mime_part_openpgp_verify
g_mime_part_set_content
g_mime_part_set_content_description
g_mime_part_set_content_encoding
g_mime_part_set_content_id
g_mime_part_set_content_location
g_mime_part_set_content_md5
g_mime_part_set_filename
g_mime_part_set_openpgp_data
g_mime_part_verify_content_md5
g_mime_pkcs7_context_get_type
g_mime_pkcs7_context_new
g_mime_references_append
g_mime_references_clear
g_mime_references_copy
g_mime_references_free
g_mime_references_get_message_id
g_mime_references_get_type
g_mime_references_length
g_mime_references_new
g_mime_references_parse
g_mime_references_set_message_id
g_mime_shutdown
g_mime_signature_get_certificate
g_mime_signature_get_created
g_mime_signature_get_expires
g_mime_signature_get_status
g_mime_signature_get_type
g_mime_signature_list_add
g_mime_signature_list_clear
g_mime_signature_list_contains
g_mime_signature_list_get_signature
g_mime_signature_list_get_type
g_mime_signature_list_index_of
g_mime_signature_list_insert
g_mime_signature_list_length
g_mime_signature_list_new
g_mime_signature_list_remove
g_mime_signature_list_remove_at
g_mime_signature_list_set_signature
g_mime_signature_new
g_mime_signature_set_certificate
g_mime_signature_set_created
g_mime_signature_set_expires
g_mime_signature_set_status
g_mime_stream_buffer_get_type
g_mime_stream_buffer_gets
g_mime_stream_buffer_new
g_mime_stream_buffer_readln
g_mime_stream_cat_add_source
g_mime_stream_cat_get_type
g_mime_stream_cat_new
g_mime_stream_close
g_mime_stream_construct
g_mime_stream_eos
g_mime_stream_file_get_owner
g_mime_stream_file_get_type
g_mime_stream_file_new
g_mime_stream_file_new_with_bounds
g_mime_stream_file_open
g_mime_stream_file_set_owner
g_mime_stream_filter_add
g_mime_stream_filter_get_owner
g_mime_stream_filter_get_type
g_mime_stream_filter_new
g_mime_stream_filter_remove
g_mime_stream_filter_set_owner
g_mime_stream_flush
g_mime_stream_fs_get_owner
g_mime_stream_fs_get_type
g_mime_stream_fs_new
g_mime_stream_fs_new_with_bounds
g_mime_stream_fs_open
g_mime_stream_fs_set_owner
g_mime_stream_get_type
g_mime_stream_gio_get_owner
g_mime_stream_gio_get_type
g_mime_stream_gio_new
g_mime_stream_gio_new_with_bounds
g_mime_stream_gio_set_owner
g_mime_stream_length
g_mime_stream_mem_get_byte_array
g_mime_stream_mem_get_owner
g_mime_stream_mem_get_type
g_mime_stream_mem_new
g_mime_stream_mem_new_with_buffer
g_mime_stream_mem_new_with_byte_array
g_mime_stream_mem_set_byte_array
g_mime_stream_mem_set_owner
g_mime_stream_mmap_get_owner
g_mime_stream_mmap_get_type
g_mime_stream_mmap_new
g_mime_stream_mmap_new_with_bounds
g_mime_stream_mmap_set_owner
g_mime_stream_null_set_count_newlines
g_mime_stream_null_get_count_newlines
g_mime_stream_null_get_type
g_mime_stream_null_new
g_mime_stream_pipe_get_owner
g_mime_stream_pipe_get_type
g_mime_stream_pipe_new
g_mime_stream_pipe_set_owner
g_mime_stream_printf
g_mime_stream_read
g_mime_stream_reset
g_mime_stream_seek
g_mime_stream_set_bounds
g_mime_stream_substream
g_mime_stream_tell
g_mime_stream_write
g_mime_stream_write_string
g_mime_stream_write_to_stream
g_mime_stream_writev
g_mime_text_part_get_charset
g_mime_text_part_get_text
g_mime_text_part_get_type
g_mime_text_part_new
g_mime_text_part_new_with_subtype
g_mime_text_part_set_charset
g_mime_text_part_set_text
g_mime_utils_best_encoding
g_mime_utils_decode_8bit
g_mime_utils_decode_message_id
g_mime_utils_generate_message_id
g_mime_utils_header_decode_date
g_mime_utils_header_decode_phrase
g_mime_utils_header_decode_text
g_mime_utils_header_encode_phrase
g_mime_utils_header_encode_text
g_mime_utils_header_format_date
g_mime_utils_header_printf
g_mime_utils_header_unfold
g_mime_utils_quote_string
g_mime_utils_structured_header_fold
g_mime_utils_text_is_8bit
g_mime_utils_unquote_string
g_mime_utils_unstructured_header_fold
g_mime_ydecode_step
g_mime_yencode_close
g_mime_yencode_step
gmime_binary_age
gmime_error_quark
gmime_gpgme_error_quark
gmime_interface_age
gmime_major_version
gmime_micro_version
gmime_minor_version
internet_address_get_charset
internet_address_get_name
internet_address_get_type
internet_address_group_add_member
internet_address_group_get_members
internet_address_group_get_type
internet_address_group_new
internet_address_group_set_members
internet_address_list_add
internet_address_list_append
internet_address_list_clear
internet_address_list_contains
internet_address_list_encode
internet_address_list_get_address
internet_address_list_get_type
internet_address_list_index_of
internet_address_list_insert
internet_address_list_length
internet_address_list_new
internet_address_list_parse
internet_address_list_prepend
internet_address_list_remove
internet_address_list_remove_at
internet_address_list_set_address
internet_address_list_to_string
internet_address_mailbox_get_addr
internet_address_mailbox_get_idn_addr
internet_address_mailbox_get_type
internet_address_mailbox_new
internet_address_mailbox_set_addr
internet_address_set_charset
internet_address_set_name
internet_address_to_string
g_mime_autocrypt_header_new
g_mime_autocrypt_header_new_from_string
g_mime_autocrypt_header_get_address
g_mime_autocrypt_header_set_address
g_mime_autocrypt_header_set_address_from_string
g_mime_autocrypt_header_get_address_as_string
g_mime_autocrypt_header_get_prefer_encrypt
g_mime_autocrypt_header_set_prefer_encrypt
g_mime_autocrypt_header_get_keydata
g_mime_autocrypt_header_set_keydata
g_mime_autocrypt_header_get_effective_date
g_mime_autocrypt_header_set_effective_date
g_mime_autocrypt_header_is_complete
g_mime_autocrypt_header_to_string
g_mime_autocrypt_header_compare
g_mime_autocrypt_header_clone
g_mime_autocrypt_header_list_new
g_mime_autocrypt_header_list_add_missing_addresses
g_mime_autocrypt_header_list_add
g_mime_autocrypt_header_list_get_count
g_mime_autocrypt_header_list_get_header_at
g_mime_autocrypt_header_list_get_header_for_address
g_mime_autocrypt_header_list_remove_incomplete
g_mime_autocrypt_header_get_type
g_mime_autocrypt_header_list_get_type
